-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: AVS wrapper to make a MM slave ports for ETH available as conduit
-- Description:
-- Remark:
-- . The avs2_eth_coe_hw.tcl determines the read latency per port

LIBRARY IEEE, common_lib, tech_tse_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE tech_tse_lib.tech_tse_pkg.ALL;
USE work.eth_pkg.ALL;


ENTITY avs2_eth_coe IS
  PORT (
    ----------------------------------------------------------------------------
    -- MM side
    
    -- Clock interface
    csi_mm_reset               : IN  STD_LOGIC;
    csi_mm_clk                 : IN  STD_LOGIC;
    
    -- TSE MAC
    mms_tse_address            : IN  STD_LOGIC_VECTOR(c_tech_tse_byte_addr_w-1 DOWNTO 0);
    mms_tse_write              : IN  STD_LOGIC;
    mms_tse_read               : IN  STD_LOGIC;
    mms_tse_writedata          : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    mms_tse_readdata           : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- read latency is 0
    mms_tse_waitrequest        : OUT STD_LOGIC;                              -- necessary because read latency is 0
    
    -- ETH registers
    mms_reg_address            : IN  STD_LOGIC_VECTOR(c_eth_reg_addr_w-1 DOWNTO 0);
    mms_reg_write              : IN  STD_LOGIC;
    mms_reg_read               : IN  STD_LOGIC;
    mms_reg_writedata          : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    mms_reg_readdata           : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- read latency is 1
    
    -- ETH packet RAM
    mms_ram_address            : IN  STD_LOGIC_VECTOR(c_eth_ram_addr_w-1 DOWNTO 0);
    mms_ram_write              : IN  STD_LOGIC;
    mms_ram_read               : IN  STD_LOGIC;
    mms_ram_writedata          : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    mms_ram_readdata           : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- read latency is 2
    
    -- Interrupt Sender interface
    ins_interrupt_irq          : OUT STD_LOGIC;  -- relates to the ETH registers port
    
    ----------------------------------------------------------------------------
    -- User side
    
    -- Clock interface
    coe_reset_export           : OUT STD_LOGIC;
    coe_clk_export             : OUT STD_LOGIC;
    
    -- TSE MAC
    coe_tse_address_export     : OUT STD_LOGIC_VECTOR(c_tech_tse_byte_addr_w-1 DOWNTO 0);
    coe_tse_write_export       : OUT STD_LOGIC;
    coe_tse_read_export        : OUT STD_LOGIC;
    coe_tse_writedata_export   : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    coe_tse_readdata_export    : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    coe_tse_waitrequest_export : IN  STD_LOGIC;
    
    -- ETH registers
    coe_reg_address_export     : OUT STD_LOGIC_VECTOR(c_eth_reg_addr_w-1 DOWNTO 0);
    coe_reg_write_export       : OUT STD_LOGIC;
    coe_reg_read_export        : OUT STD_LOGIC;
    coe_reg_writedata_export   : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    coe_reg_readdata_export    : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    
    -- ETH packet RAM
    coe_ram_address_export     : OUT STD_LOGIC_VECTOR(c_eth_ram_addr_w-1 DOWNTO 0);
    coe_ram_write_export       : OUT STD_LOGIC;
    coe_ram_read_export        : OUT STD_LOGIC;
    coe_ram_writedata_export   : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    coe_ram_readdata_export    : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    
    -- Interrupt Sender interface
    coe_irq_export             : IN  STD_LOGIC
  );
END avs2_eth_coe;


ARCHITECTURE wrap OF avs2_eth_coe IS
BEGIN

  ------------------------------------------------------------------------------
  -- Wires
  
  -- Clock interface
  coe_reset_export         <= csi_mm_reset;
  coe_clk_export           <= csi_mm_clk;
  
  -- TSE MAC
  coe_tse_address_export   <= mms_tse_address;
  coe_tse_write_export     <= mms_tse_write;
  coe_tse_read_export      <= mms_tse_read;
  coe_tse_writedata_export <= mms_tse_writedata;      
  mms_tse_readdata         <= coe_tse_readdata_export;
  mms_tse_waitrequest      <= coe_tse_waitrequest_export;
                                                            
  -- ETH registers
  coe_reg_address_export   <= mms_reg_address;
  coe_reg_write_export     <= mms_reg_write;
  coe_reg_read_export      <= mms_reg_read;
  coe_reg_writedata_export <= mms_reg_writedata;
  mms_reg_readdata         <= coe_reg_readdata_export;
                                                            
  -- ETH packet RAM
  coe_ram_address_export   <= mms_ram_address;
  coe_ram_write_export     <= mms_ram_write;
  coe_ram_read_export      <= mms_ram_read;
  coe_ram_writedata_export <= mms_ram_writedata;
  mms_ram_readdata         <= coe_ram_readdata_export;
                                                            
  -- Interrupt Sender interface
  ins_interrupt_irq        <= coe_irq_export;

END wrap;
