-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program IS free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program IS distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.eth_pkg.ALL;

-- Purpose:
-- Description:
--   Strip IPv4 headers to 20 bytes
--   Set ihl to 5 (5 * 4 = 20 bytes)
--   Ajust IPv4 Packet Total Length field
--   IP Header Checksum is set to 0x0000 and should be ignored
--   The checksum could be updated using https://tools.ietf.org/html/rfc1624
-- Remarks:
--  . c_this_snk_latency = 1
--  . c_this_src_latency = 1

  ------------------------------------------------------------------------------
  -- IPv4 Packet
  --
  --  0       3 4     7 8            15 16   18 19                        31  wi
  -- |----------------------------------------------------------------------|
  -- | Version |  ihl  |    Services    |    IPv4 Packet Total Length       |  4
  -- |----------------------------------------------------------------------|
  -- |       Identification             | Flags |    Fragment Offset        |  5
  -- |----------------------------------------------------------------------|
  -- |     TTL         |    Protocol    |      Header Checksum              |  6
  -- |----------------------------------------------------------------------|
  -- |              Source IP Address                                       |  7
  -- |----------------------------------------------------------------------|
  -- |              Destination IP Address                                  |  8
  -- |----------------------------------------------------------------------|
  -- |              Options ...                                             |
  -- |----------------------------------------------------------------------|


ENTITY eth_ihl_to_20 IS
  GENERIC (
    incoming_ihl   : NATURAL := 20
  );
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    -- Streaming Sink
    snk_in         : IN  t_dp_sosi;
    snk_out        : OUT t_dp_siso;
    
    -- Streaming Source
    src_in         : IN  t_dp_siso;
    src_out        : OUT t_dp_sosi
  );
END eth_ihl_to_20;


ARCHITECTURE rtl OF eth_ihl_to_20 IS

  SIGNAL i_src_out        : t_dp_sosi;
  
  TYPE state_type IS (Idle, Eth_DestMAC0, Eth_DestMAC1, Eth_SrcMAC0, Eth_SrcMAC1, IPv4_lengths, IPv4_ID, IPv4_TTL, IPv4_SrcIP, IPv4_DestIP, IPv4_Options, IPv4_Payload);
  SIGNAL state   : state_type;
  
  SIGNAL ihl : unsigned(c_network_ip_header_length_w-1 downto 0);
  
BEGIN

  src_out <= i_src_out;
  --i_src_out <= snk_in;

  -- Pass on frame level flow control
  snk_out.xon <= src_in.xon;
  
  -- No change in ready latency, c_this_snk_latency = c_this_src_latency
  snk_out.ready <= src_in.ready;
  

  PROCESS(clk, rst)
  BEGIN
    IF rst = '1' THEN 
      state <= Idle;
      ihl   <= TO_UNSIGNED(0,ihl'LENGTH);
    ELSIF (rising_edge(clk)) THEN
      i_src_out <= snk_in;
      CASE state IS
        WHEN Idle=>
          IF snk_in.sop = '1' and snk_in.valid = '1' THEN 
            state <= Eth_DestMAC0;
          END IF;
          
        WHEN Eth_DestMAC0=>
          IF snk_in.valid = '1' THEN 
            state <= Eth_DestMAC1;
          END IF;
        
        WHEN Eth_DestMAC1  =>
          IF snk_in.valid = '1' THEN 
            state <= Eth_SrcMAC0;
          END IF;
        
        WHEN Eth_SrcMAC0  =>
          IF snk_in.valid = '1' THEN 
            IF snk_in.data(c_network_eth_type_slv'RANGE) = TO_UVEC(c_network_eth_type_ip, c_network_eth_type_w) THEN -- if IPv4 frame
              state <= Eth_SrcMAC1;
            ELSE
              state <= Idle;
            END IF;
          END IF;
          
        WHEN Eth_SrcMAC1  =>
          IF snk_in.valid = '1' THEN
            IF snk_in.data(27 downto 24) = TO_UVEC(c_network_ip_header_length, c_network_ip_header_length_w) THEN  -- if ihl = 5 words = 20 bytes, nothing to do
              state <= Idle;
            ELSE
              state <= IPv4_lengths;
              i_src_out.data(27 downto 24) <= TO_UVEC(c_network_ip_header_length, c_network_ip_header_length_w);  -- make IHL = 5 words = 20 bytes
              i_src_out.data(c_network_ip_total_length_w-1 downto 0) <= std_logic_vector(
                                             unsigned(snk_in.data(c_network_ip_total_length_w-1 downto 0)) -
                                           ((unsigned(snk_in.data(27 downto 24)) - to_unsigned(c_network_ip_header_length, c_network_ip_header_length_w)) & "00")
                                                                                        );                        -- correct Total Length = input Total Length - (input IHL - fixed IHL),
              ihl <= unsigned(snk_in.data(27 downto 24));  -- save input IHL
            END IF;
          END IF;
        
        WHEN IPv4_lengths  =>
          IF snk_in.valid = '1' THEN
            state <= IPv4_ID;
            ihl <= ihl - 1;
          END IF;
      
        WHEN IPv4_ID  =>
          IF snk_in.valid = '1' THEN
            state <= IPv4_TTL;
            i_src_out.data(c_network_ip_header_checksum_w-1 downto 0) <= TO_UVEC(0, c_network_ip_header_checksum_w);
            ihl <= ihl - 1;
          END IF;
        
        WHEN IPv4_TTL  =>
          IF snk_in.valid = '1' THEN
            state <= IPv4_SrcIP;
            ihl <= ihl - 1;
          END IF;
        
        WHEN IPv4_SrcIP  =>
          IF snk_in.valid = '1' THEN
            state <= IPv4_DestIP;
            ihl <= ihl - 1;
          END IF;          
        
        WHEN IPv4_DestIP  =>
          IF snk_in.valid = '1' THEN
            i_src_out.valid <= '0';    -- do not allow option words to get through this module
            state <= IPv4_Options;
            ihl <= ihl - 1;
            IF ihl = 2 THEN
              state <= IPv4_Payload;
            END IF;
          END IF;
          
        WHEN IPv4_Options  =>
          IF snk_in.valid = '1' THEN
            i_src_out.valid <= '0';    -- do not allow option words to get through this module
            ihl <= ihl - 1;
            IF ihl = 2 THEN
              state <= IPv4_Payload;
            END IF;
          END IF;
          
        WHEN IPv4_Payload =>
          IF snk_in.eop = '1' and snk_in.valid = '1' THEN
            state <= Idle;
          END IF;
          
        WHEN others =>
          state <= Idle;
      END CASE;
    END IF;
  end PROCESS;
  
END rtl;
